@@ -51,7 +51,7 @@ void cm_InitConn(void)
 	HKEY parmKey;
         
     if (osi_Once(&once)) {
-		lock_InitializeRWLock(&cm_connLock, "connection global lock");
+	lock_InitializeRWLock(&cm_connLock, "connection global lock");
 
         /* keisa - read timeout value for lanmanworkstation  service.
          * jaltman - as per 
@@ -60,29 +60,30 @@ void cm_InitConn(void)
          * I believe that the default should not be short.  Instead, we should wait until
          * RX times out before reporting a timeout to the SMB client.
          */
-		code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, LANMAN_WKS_PARAM_KEY,
-                            0, KEY_QUERY_VALUE, &parmKey);
-		if (code == ERROR_SUCCESS)
+	code = RegOpenKeyEx(HKEY_LOCAL_MACHINE, LANMAN_WKS_PARAM_KEY,
+			    0, KEY_QUERY_VALUE, &parmKey);
+	if (code == ERROR_SUCCESS)
         {
-		    DWORD dummyLen = sizeof(sessTimeout);
-		    code = RegQueryValueEx(parmKey, LANMAN_WKS_SESSION_TIMEOUT, NULL, NULL, 
-                                   (BYTE *) &sessTimeout, &dummyLen);
-		    if (code == ERROR_SUCCESS)
+	    DWORD dummyLen = sizeof(sessTimeout);
+	    code = RegQueryValueEx(parmKey, LANMAN_WKS_SESSION_TIMEOUT, NULL, NULL, 
+				   (BYTE *) &sessTimeout, &dummyLen);
+	    if (code == ERROR_SUCCESS)
             {
-                afsi_log("lanmanworkstation : SessTimeout %d", sessTimeout);
                 RDRtimeout = sessTimeout;
-                if ( ConnDeadtimeout < RDRtimeout + 15 ) {
-                    ConnDeadtimeout = RDRtimeout + 15;
-                    afsi_log("ConnDeadTimeout increased to %d", ConnDeadtimeout);
-                }
-                if ( HardDeadtimeout < 2 * ConnDeadtimeout ) {
-                    HardDeadtimeout = 2 * ConnDeadtimeout;
-                    afsi_log("HardDeadTimeout increased to %d", HardDeadtimeout);
-                }
-            }
+	    }
         }
 
-        osi_EndOnce(&once);
+	afsi_log("lanmanworkstation : SessTimeout %d", sessTimeout);
+	if ( ConnDeadtimeout < RDRtimeout + 15 ) {
+	    ConnDeadtimeout = RDRtimeout + 15;
+	    afsi_log("ConnDeadTimeout increased to %d", ConnDeadtimeout);
+	}
+	if ( HardDeadtimeout < 2 * ConnDeadtimeout ) {
+	    HardDeadtimeout = 2 * ConnDeadtimeout;
+	    afsi_log("HardDeadTimeout increased to %d", HardDeadtimeout);
+	}
+
+	osi_EndOnce(&once);
     }
 }
 
@@ -90,7 +91,7 @@ void cm_InitReq(cm_req_t *reqp)
 {
 	memset((char *)reqp, 0, sizeof(cm_req_t));
 #ifndef DJGPP
-	reqp->startTime = GetCurrentTime();
+	reqp->startTime = GetTickCount();
 #else
         gettimeofday(&reqp->startTime, NULL);
 #endif
@@ -143,14 +144,16 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
            cm_serverRef_t * serversp,
            cm_callbackRequest_t *cbrp, long errorCode)
 {
-    cm_server_t *serverp = 0;
-    cm_serverRef_t **serverspp = 0;
+    cm_server_t *serverp = NULL;
+    cm_serverRef_t **serverspp = NULL;
     cm_serverRef_t *tsrp;
+    cm_cell_t  *cellp = NULL;
     cm_ucell_t *ucellp;
     int retry = 0;
     int free_svr_list = 0;
     int dead_session;
     long timeUsed, timeLeft;
+    long code;
         
     osi_Log2(afsd_logp, "cm_Analyze connp 0x%x, code 0x%x",
              (long) connp, errorCode);
@@ -182,24 +185,33 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
     if (reqp->flags & CM_REQ_NORETRY)
         goto out;
 
-    /* if timeout - check that it did not exceed the SMB timeout
+    /* if timeout - check that it did not exceed the HardDead timeout
      * and retry */
     
     /* timeleft - get if from reqp the same way as cmXonnByMServers does */
 #ifndef DJGPP
-    timeUsed = (GetCurrentTime() - reqp->startTime) / 1000;
+    timeUsed = (GetTickCount() - reqp->startTime) / 1000;
 #else
     gettimeofday(&now, NULL);
     timeUsed = sub_time(now, reqp->startTime) / 1000;
 #endif
 	    
     /* leave 5 seconds margin for sleep */
-    timeLeft = RDRtimeout - timeUsed;
+    timeLeft = HardDeadtimeout - timeUsed;
 
     if (errorCode == CM_ERROR_TIMEDOUT) {
         if (timeLeft > 5 ) {
             thrd_Sleep(3000);
-            cm_CheckServers(CM_FLAG_CHECKDOWNSERVERS, NULL);
+            if (cellp == NULL && serverp)
+                cellp = serverp->cellp;
+            if (cellp == NULL && serversp) {
+                struct cm_serverRef * refp;
+                for ( refp=serversp ; cellp == NULL && refp != NULL; refp=refp->next) {
+                    if ( refp->server )
+                        cellp = refp->server->cellp;
+                }
+            }
+            cm_CheckServers(CM_FLAG_CHECKDOWNSERVERS, cellp);
             retry = 1;
         }
     } 
@@ -211,8 +223,8 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
      * until the timeout period expires.
      */
     else if (errorCode == CM_ERROR_NOSUCHVOLUME) {
+	osi_Log0(afsd_logp, "cm_Analyze passed CM_ERROR_NOSUCHVOLUME.");
         if (timeLeft > 7) {
-            osi_Log0(afsd_logp, "cm_Analyze passed CM_ERROR_NOSUCHVOLUME.");
             thrd_Sleep(5000);
             
             retry = 1;
@@ -222,48 +234,65 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
         }
     }
 
+    else if (errorCode == CM_ERROR_ALLDOWN) {
+	osi_Log0(afsd_logp, "cm_Analyze passed CM_ERROR_ALLDOWN.");
+	/* Servers marked DOWN will be restored by the background daemon
+	 * thread as they become available.
+	 */
+    }
+
     else if (errorCode == CM_ERROR_ALLOFFLINE) {
         if (timeLeft > 7) {
             osi_Log0(afsd_logp, "cm_Analyze passed CM_ERROR_ALLOFFLINE.");
             thrd_Sleep(5000);
-            /* cm_ForceUpdateVolume marks all servers as non_busy */
-            /* No it doesn't and it won't do anything if all of the 
-             * the servers are marked as DOWN.  So clear the DOWN
-             * flag and reset the busy state as well.
-             */
-            if (!serversp) {
-                cm_GetServerList(fidp, userp, reqp, &serverspp);
-                serversp = *serverspp;
-                free_svr_list = 1;
-            }
-            if (serversp) {
-                lock_ObtainWrite(&cm_serverLock);
-                for (tsrp = serversp; tsrp; tsrp=tsrp->next) {
-                    tsrp->server->flags &= ~CM_SERVERFLAG_DOWN;
-                    if (tsrp->status == busy)
-                        tsrp->status = not_busy;
-                }
-                lock_ReleaseWrite(&cm_serverLock);
-                if (free_svr_list) {
-                    cm_FreeServerList(&serversp);
-                    *serverspp = serversp;
-                }
-                retry = 1;
-            }
+            
+	    if (fidp) {	/* Not a VLDB call */
+		if (!serversp) {
+		    code = cm_GetServerList(fidp, userp, reqp, &serverspp);
+		    if (code == 0) {
+			serversp = *serverspp;
+			free_svr_list = 1;
+		    }
+		}
+		if (serversp) {
+		    lock_ObtainWrite(&cm_serverLock);
+		    for (tsrp = serversp; tsrp; tsrp=tsrp->next)
+			tsrp->status = not_busy;
+		    lock_ReleaseWrite(&cm_serverLock);
+		    if (free_svr_list) {
+			cm_FreeServerList(&serversp);
+			*serverspp = serversp;
+		    }
+		    retry = 1;
+		}
 
-            if (fidp != NULL)   /* Not a VLDB call */
                 cm_ForceUpdateVolume(fidp, userp, reqp);
+	    } else { /* VLDB call */
+		if (serversp) {
+		    lock_ObtainWrite(&cm_serverLock);
+		    for (tsrp = serversp; tsrp; tsrp=tsrp->next)
+			tsrp->status = not_busy;
+		    lock_ReleaseWrite(&cm_serverLock);
+		    if (free_svr_list) {
+			cm_FreeServerList(&serversp);
+			*serverspp = serversp;
+		    }
+		}
+	    }	
         }
     }
 
     /* if all servers are busy, mark them non-busy and start over */
     else if (errorCode == CM_ERROR_ALLBUSY) {
+	osi_Log0(afsd_logp, "cm_Analyze passed CM_ERROR_ALLBUSY.");
         if (timeLeft > 7) {
             thrd_Sleep(5000);
             if (!serversp) {
-                cm_GetServerList(fidp, userp, reqp, &serverspp);
-                serversp = *serverspp;
-                free_svr_list = 1;
+                code = cm_GetServerList(fidp, userp, reqp, &serverspp);
+                if (code == 0) {
+                    serversp = *serverspp;
+                    free_svr_list = 1;
+                }
             }
             lock_ObtainWrite(&cm_serverLock);
             for (tsrp = serversp; tsrp; tsrp=tsrp->next) {
@@ -282,9 +311,11 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
     /* special codes:  VBUSY and VRESTARTING */
     else if (errorCode == VBUSY || errorCode == VRESTARTING) {
         if (!serversp) {
-            cm_GetServerList(fidp, userp, reqp, &serverspp);
-            serversp = *serverspp;
-            free_svr_list = 1;
+            code = cm_GetServerList(fidp, userp, reqp, &serverspp);
+            if (code == 0) {
+                serversp = *serverspp;
+                free_svr_list = 1;
+            }
         }
         lock_ObtainWrite(&cm_serverLock);
         for (tsrp = serversp; tsrp; tsrp=tsrp->next) {
@@ -303,38 +334,63 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
     }
 
     /* special codes:  missing volumes */
-    else if (errorCode == VNOVOL || errorCode == VMOVED || errorCode == VOFFLINE
-         || errorCode == VSALVAGE || errorCode == VNOSERVICE) 
+    else if (errorCode == VNOVOL || errorCode == VMOVED || errorCode == VOFFLINE ||
+             errorCode == VSALVAGE || errorCode == VNOSERVICE || errorCode == VIO) 
     {       
-        /* Log server being offline for this volume */
-        osi_Log4(afsd_logp, "cm_Analyze found server %d.%d.%d.%d marked offline for a volume",
-                  ((serverp->addr.sin_addr.s_addr & 0xff)),
-                  ((serverp->addr.sin_addr.s_addr & 0xff00)>> 8),
-                  ((serverp->addr.sin_addr.s_addr & 0xff0000)>> 16),
-                  ((serverp->addr.sin_addr.s_addr & 0xff000000)>> 24));
-        /* Create Event Log message */ 
-        {
-            HANDLE h;
-            char *ptbuf[1];
-            char s[100];
-            h = RegisterEventSource(NULL, AFS_DAEMON_EVENT_NAME);
-            sprintf(s, "cm_Analyze: Server %d.%d.%d.%d reported volume %d as missing.",
-                     ((serverp->addr.sin_addr.s_addr & 0xff)),
-                     ((serverp->addr.sin_addr.s_addr & 0xff00)>> 8),
-                     ((serverp->addr.sin_addr.s_addr & 0xff0000)>> 16),
-                     ((serverp->addr.sin_addr.s_addr & 0xff000000)>> 24),
-                     fidp->volume);
-            ptbuf[0] = s;
-            ReportEvent(h, EVENTLOG_WARNING_TYPE, 0, 1009, NULL,
-                         1, 0, ptbuf, NULL);
-            DeregisterEventSource(h);
+#ifndef DJGPP
+        HANDLE h;
+        char *ptbuf[1];
+        char s[100];
+#endif
+        char addr[16];
+        char *format;
+
+        switch ( errorCode ) {
+        case VNOVOL:
+            format = "Server %s reported volume %d as not attached.";
+            break;
+        case VMOVED:
+            format = "Server %s reported volume %d as moved.";
+            break;
+        case VOFFLINE:
+            format = "Server %s reported volume %d as offline.";
+            break;
+        case VSALVAGE:
+            format = "Server %s reported volume %d as needs salvage.";
+            break;
+        case VNOSERVICE:
+            format = "Server %s reported volume %d as not in service.";
+            break;
+        case VIO:
+            format = "Server %s reported volume %d as temporarily unaccessible.";
+            break;
         }
 
+        /* Log server being offline for this volume */
+        sprintf(addr, "%d.%d.%d.%d", 
+                 ((serverp->addr.sin_addr.s_addr & 0xff)),
+                 ((serverp->addr.sin_addr.s_addr & 0xff00)>> 8),
+                 ((serverp->addr.sin_addr.s_addr & 0xff0000)>> 16),
+                 ((serverp->addr.sin_addr.s_addr & 0xff000000)>> 24)); 
+        osi_Log2(afsd_logp, format, osi_LogSaveString(afsd_logp,addr), fidp->volume);
+        
+#ifndef DJGPP
+        /* Create Event Log message */
+        h = RegisterEventSource(NULL, AFS_DAEMON_EVENT_NAME);
+        sprintf(s, format, addr, fidp->volume);
+        ptbuf[0] = s;
+        ReportEvent(h, EVENTLOG_WARNING_TYPE, 0, 1009, NULL,
+                     1, 0, ptbuf, NULL);
+        DeregisterEventSource(h);
+#endif
+
         /* Mark server offline for this volume */
         if (!serversp) {
-            cm_GetServerList(fidp, userp, reqp, &serverspp);
-            serversp = *serverspp;
-            free_svr_list = 1;
+            code = cm_GetServerList(fidp, userp, reqp, &serverspp);
+            if (code == 0) {
+                serversp = *serverspp;
+                free_svr_list = 1;
+            }
         }
         for (tsrp = serversp; tsrp; tsrp=tsrp->next) {
             if (tsrp->server == serverp)
@@ -346,6 +402,36 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
         }
         if ( timeLeft > 2 )
             retry = 1;
+    } else if ( errorCode == VNOVNODE ) {
+	if ( fidp ) {
+	    cm_scache_t * scp;
+	    osi_Log4(afsd_logp, "cm_Analyze passed VNOVNODE cell %u vol %u vn %u uniq %u.",
+		      fidp->cell, fidp->volume, fidp->vnode, fidp->unique);
+
+	    scp = cm_FindSCache(fidp);
+	    if (scp) {
+		cm_scache_t *pscp = NULL;
+		
+		if (scp->fileType != CM_SCACHETYPE_DIRECTORY)
+		    pscp = cm_FindSCacheParent(scp);
+
+		lock_ObtainWrite(&cm_scacheLock);
+		cm_RecycleSCache(scp, CM_SCACHE_RECYCLEFLAG_DESTROY_BUFFERS);
+		cm_ReleaseSCacheNoLock(scp);
+		lock_ReleaseWrite(&cm_scacheLock);
+
+		if (pscp) {
+		    if (pscp->cbExpires > 0 && pscp->cbServerp != NULL) {
+			lock_ObtainMutex(&pscp->mx);
+			cm_DiscardSCache(pscp);
+			lock_ReleaseMutex(&pscp->mx);
+		    }
+		    cm_ReleaseSCache(pscp);
+		}
+	    }
+	} else {
+	    osi_Log0(afsd_logp, "cm_Analyze passed VNOVNODE unknown fid.");
+	}
     }
 
     /* RX codes */
@@ -375,10 +461,12 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
         lock_ObtainMutex(&serverp->mx);
         serverp->flags |= CM_SERVERFLAG_DOWN;
         lock_ReleaseMutex(&serverp->mx);
+	cm_ForceNewConnections(serverp);
         if ( timeLeft > 2 )
             retry = 1;
     }
-    else if (errorCode == RXKADEXPIRED) {
+    else if (errorCode == RXKADEXPIRED || 
+             errorCode == RXKADBADTICKET) {
         if (!dead_session) {
             lock_ObtainMutex(&userp->mx);
             ucellp = cm_GetUCell(userp, serverp->cellp);
@@ -393,8 +481,96 @@ cm_Analyze(cm_conn_t *connp, cm_user_t *userp, cm_req_t *reqp,
                 retry = 1;
         }
     } else {
-        if (errorCode)
-            osi_Log1(afsd_logp, "cm_Analyze: ignoring error code 0x%x", errorCode);
+        if (errorCode) {
+            char * s = "unknown error";
+            switch ( errorCode ) {
+            case RXKADINCONSISTENCY: s = "RXKADINCONSISTENCY"; break;
+            case RXKADPACKETSHORT  : s = "RXKADPACKETSHORT";   break;
+            case RXKADLEVELFAIL    : s = "RXKADLEVELFAIL";     break;
+            case RXKADTICKETLEN    : s = "RXKADTICKETLEN";     break;
+            case RXKADOUTOFSEQUENCE: s = "RXKADOUTOFSEQUENCE"; break;
+            case RXKADNOAUTH       : s = "RXKADNOAUTH";        break;
+            case RXKADBADKEY       : s = "RXKADBADKEY";        break;
+            case RXKADBADTICKET    : s = "RXKADBADTICKET";     break;
+            case RXKADUNKNOWNKEY   : s = "RXKADUNKNOWNKEY";    break;
+            case RXKADEXPIRED      : s = "RXKADEXPIRED";       break;
+            case RXKADSEALEDINCON  : s = "RXKADSEALEDINCON";   break;
+            case RXKADDATALEN      : s = "RXKADDATALEN";       break;
+            case RXKADILLEGALLEVEL : s = "RXKADILLEGALLEVEL";  break;
+            case VSALVAGE          : s = "VSALVAGE";           break;
+            case VNOVNODE          : s = "VNOVNODE";           break;
+            case VNOVOL            : s = "VNOVOL";             break;
+            case VVOLEXISTS        : s = "VVOLEXISTS";         break;
+            case VNOSERVICE        : s = "VNOSERVICE";         break;
+            case VOFFLINE          : s = "VOFFLINE";           break;
+            case VONLINE           : s = "VONLINE";            break;
+            case VDISKFULL         : s = "VDISKFULL";          break;
+            case VOVERQUOTA        : s = "VOVERQUOTA";         break;
+            case VBUSY             : s = "VBUSY";              break;
+            case VMOVED            : s = "VMOVED";             break;
+            case VIO               : s = "VIO";                break;
+            case VRESTRICTED       : s = "VRESTRICTED";        break;
+            case VRESTARTING       : s = "VRESTARTING";        break;
+            case VREADONLY         : s = "VREADONLY";          break;
+            case EAGAIN            : s = "EAGAIN";             break;
+            case EACCES            : s = "EACCES";             break;
+	    case ENOENT            : s = "ENOENT";             break;
+            case CM_ERROR_NOSUCHCELL	    : s = "CM_ERROR_NOSUCHCELL";         break; 			
+            case CM_ERROR_NOSUCHVOLUME	    : s = "CM_ERROR_NOSUCHVOLUME";       break; 			
+            case CM_ERROR_TIMEDOUT	    : s = "CM_ERROR_TIMEDOUT";           break; 		
+            case CM_ERROR_RETRY		    : s = "CM_ERROR_RETRY";              break; 
+            case CM_ERROR_NOACCESS	    : s = "CM_ERROR_NOACCESS";           break; 
+            case CM_ERROR_NOSUCHFILE	    : s = "CM_ERROR_NOSUCHFILE";         break; 			
+            case CM_ERROR_STOPNOW	    : s = "CM_ERROR_STOPNOW";            break; 			
+            case CM_ERROR_TOOBIG	    : s = "CM_ERROR_TOOBIG";             break; 				
+            case CM_ERROR_INVAL		    : s = "CM_ERROR_INVAL";              break; 				
+            case CM_ERROR_BADFD		    : s = "CM_ERROR_BADFD";              break; 				
+            case CM_ERROR_BADFDOP	    : s = "CM_ERROR_BADFDOP";            break; 			
+            case CM_ERROR_EXISTS	    : s = "CM_ERROR_EXISTS";             break; 				
+            case CM_ERROR_CROSSDEVLINK	    : s = "CM_ERROR_CROSSDEVLINK";       break; 			
+            case CM_ERROR_BADOP		    : s = "CM_ERROR_BADOP";              break; 				
+            case CM_ERROR_BADPASSWORD       : s = "CM_ERROR_BADPASSWORD";        break;         
+            case CM_ERROR_NOTDIR	    : s = "CM_ERROR_NOTDIR";             break; 				
+            case CM_ERROR_ISDIR		    : s = "CM_ERROR_ISDIR";              break; 				
+            case CM_ERROR_READONLY	    : s = "CM_ERROR_READONLY";           break; 			
+            case CM_ERROR_WOULDBLOCK	    : s = "CM_ERROR_WOULDBLOCK";         break; 			
+            case CM_ERROR_QUOTA		    : s = "CM_ERROR_QUOTA";              break; 				
+            case CM_ERROR_SPACE		    : s = "CM_ERROR_SPACE";              break; 				
+            case CM_ERROR_BADSHARENAME	    : s = "CM_ERROR_BADSHARENAME";       break; 			
+            case CM_ERROR_BADTID	    : s = "CM_ERROR_BADTID";             break; 				
+            case CM_ERROR_UNKNOWN	    : s = "CM_ERROR_UNKNOWN";            break; 			
+            case CM_ERROR_NOMORETOKENS	    : s = "CM_ERROR_NOMORETOKENS";       break; 			
+            case CM_ERROR_NOTEMPTY	    : s = "CM_ERROR_NOTEMPTY";           break; 			
+            case CM_ERROR_USESTD	    : s = "CM_ERROR_USESTD";             break; 				
+            case CM_ERROR_REMOTECONN	    : s = "CM_ERROR_REMOTECONN";         break; 			
+            case CM_ERROR_ATSYS		    : s = "CM_ERROR_ATSYS";              break; 				
+            case CM_ERROR_NOSUCHPATH	    : s = "CM_ERROR_NOSUCHPATH";         break; 			
+            case CM_ERROR_CLOCKSKEW	    : s = "CM_ERROR_CLOCKSKEW";          break; 			
+            case CM_ERROR_BADSMB	    : s = "CM_ERROR_BADSMB";             break; 				
+            case CM_ERROR_ALLBUSY	    : s = "CM_ERROR_ALLBUSY";            break; 			
+            case CM_ERROR_NOFILES	    : s = "CM_ERROR_NOFILES";            break; 			
+            case CM_ERROR_PARTIALWRITE	    : s = "CM_ERROR_PARTIALWRITE";       break; 			
+            case CM_ERROR_NOIPC		    : s = "CM_ERROR_NOIPC";              break; 				
+            case CM_ERROR_BADNTFILENAME	    : s = "CM_ERROR_BADNTFILENAME";      break; 			
+            case CM_ERROR_BUFFERTOOSMALL    : s = "CM_ERROR_BUFFERTOOSMALL";     break; 			
+            case CM_ERROR_RENAME_IDENTICAL  : s = "CM_ERROR_RENAME_IDENTICAL";   break; 		
+            case CM_ERROR_ALLOFFLINE        : s = "CM_ERROR_ALLOFFLINE";         break;          
+            case CM_ERROR_AMBIGUOUS_FILENAME: s = "CM_ERROR_AMBIGUOUS_FILENAME"; break;  
+            case CM_ERROR_BADLOGONTYPE	    : s = "CM_ERROR_BADLOGONTYPE";       break; 	    
+            case CM_ERROR_GSSCONTINUE       : s = "CM_ERROR_GSSCONTINUE";        break;         
+            case CM_ERROR_TIDIPC            : s = "CM_ERROR_TIDIPC";             break;              
+            case CM_ERROR_TOO_MANY_SYMLINKS : s = "CM_ERROR_TOO_MANY_SYMLINKS";  break;   
+            case CM_ERROR_PATH_NOT_COVERED  : s = "CM_ERROR_PATH_NOT_COVERED";   break;    
+            case CM_ERROR_LOCK_CONFLICT     : s = "CM_ERROR_LOCK_CONFLICT";      break;       
+            case CM_ERROR_SHARING_VIOLATION : s = "CM_ERROR_SHARING_VIOLATION";  break;   
+            case CM_ERROR_ALLDOWN           : s = "CM_ERROR_ALLDOWN";            break;             
+            case CM_ERROR_TOOFEWBUFS	    : s = "CM_ERROR_TOOFEWBUFS";         break; 			
+            case CM_ERROR_TOOMANYBUFS	    : s = "CM_ERROR_TOOMANYBUFS";        break; 			
+            }
+            osi_Log2(afsd_logp, "cm_Analyze: ignoring error code 0x%x (%s)", 
+                     errorCode, s);
+            retry = 0;
+        }
     }
 
     if (retry && dead_session)
@@ -416,16 +592,21 @@ long cm_ConnByMServers(cm_serverRef_t *serversp, cm_user_t *usersp,
     cm_serverRef_t *tsrp;
     cm_server_t *tsp;
     long firstError = 0;
-    int someBusy = 0, someOffline = 0, allBusy = 1, allDown = 1;
+    int someBusy = 0, someOffline = 0, allOffline = 1, allBusy = 1, allDown = 1;
     long timeUsed, timeLeft, hardTimeLeft;
 #ifdef DJGPP
     struct timeval now;
 #endif /* DJGPP */        
 
+    if (serversp == NULL) {
+	osi_Log1(afsd_logp, "cm_ConnByMServers returning 0x%x", CM_ERROR_NOSUCHVOLUME);
+	return CM_ERROR_NOSUCHVOLUME;
+    }
+
     *connpp = NULL;
 
 #ifndef DJGPP
-    timeUsed = (GetCurrentTime() - reqp->startTime) / 1000;
+    timeUsed = (GetTickCount() - reqp->startTime) / 1000;
 #else
     gettimeofday(&now, NULL);
     timeUsed = sub_time(now, reqp->startTime) / 1000;
@@ -441,13 +622,15 @@ long cm_ConnByMServers(cm_serverRef_t *serversp, cm_user_t *usersp,
         cm_GetServerNoLock(tsp);
         lock_ReleaseWrite(&cm_serverLock);
         if (!(tsp->flags & CM_SERVERFLAG_DOWN)) {
+	    allDown = 0;
             if (tsrp->status == busy) {
-                allDown = 0;
+		allOffline = 0;
                 someBusy = 1;
             } else if (tsrp->status == offline) {
-                someOffline = 1;
+		allBusy = 0;
+		someOffline = 1;
             } else {
-                allDown = 0;
+		allOffline = 0;
                 allBusy = 0;
                 code = cm_ConnByServer(tsp, usersp, connpp);
                 if (code == 0) {        /* cm_CBS only returns 0 */
@@ -474,15 +657,15 @@ long cm_ConnByMServers(cm_serverRef_t *serversp, cm_user_t *usersp,
         lock_ObtainWrite(&cm_serverLock);
         cm_PutServerNoLock(tsp);
     }   
-
     lock_ReleaseWrite(&cm_serverLock);
+
     if (firstError == 0) {
-        if (serversp == NULL)
-            firstError = CM_ERROR_NOSUCHVOLUME;
-        else if (allDown) 
-            firstError = CM_ERROR_ALLOFFLINE;
+        if (allDown) 
+            firstError = CM_ERROR_ALLDOWN;
         else if (allBusy) 
             firstError = CM_ERROR_ALLBUSY;
+	else if (allOffline || (someBusy && someOffline))
+	    firstError = CM_ERROR_ALLOFFLINE;
         else {
             osi_Log0(afsd_logp, "cm_ConnByMServers returning impossible error TIMEDOUT");
             firstError = CM_ERROR_TIMEDOUT;
@@ -597,7 +780,8 @@ long cm_ConnByServer(cm_server_t *serverp, cm_user_t *userp, cm_conn_t **connpp)
         tcp->refCount = 1;
         lock_ReleaseMutex(&tcp->mx);
     } else {
-        if ((tcp->ucgen < ucellp->gen) ||
+        if ((tcp->flags & CM_CONN_FLAG_FORCE_NEW) ||
+            (tcp->ucgen < ucellp->gen) ||
             (tcp->cryptlevel != (cryptall ? (ucellp->flags & CM_UCELLFLAG_RXKAD ? rxkad_crypt : rxkad_clear) : rxkad_clear)))
         {
             if (tcp->ucgen < ucellp->gen)
@@ -605,6 +789,7 @@ long cm_ConnByServer(cm_server_t *serverp, cm_user_t *userp, cm_conn_t **connpp)
             else
                 osi_Log0(afsd_logp, "cm_ConnByServer replace connection due to crypt change");
             lock_ObtainMutex(&tcp->mx);
+	    tcp->flags &= ~CM_CONN_FLAG_FORCE_NEW;
             rx_DestroyConnection(tcp->callp);
             cm_NewRXConnection(tcp, ucellp, serverp);
             lock_ReleaseMutex(&tcp->mx);
@@ -650,3 +835,15 @@ cm_GetRxConn(cm_conn_t *connp)
     return rxconn;
 }
 
+void cm_ForceNewConnections(cm_server_t *serverp)
+{
+    cm_conn_t *tcp;
+
+    lock_ObtainWrite(&cm_connLock);
+    for (tcp = serverp->connsp; tcp; tcp=tcp->nextp) {
+	lock_ObtainMutex(&tcp->mx);
+	tcp->flags |= CM_CONN_FLAG_FORCE_NEW;
+	lock_ReleaseMutex(&tcp->mx);
+    }
+    lock_ReleaseWrite(&cm_connLock);
+}
